from click.testing import CliRunner
import pytest
import os
import json
import time

from shipami import __version__ as VERSION
from shipami.cli import cli as shipami

runner = CliRunner()

@pytest.fixture()
def ec2():
    import boto3
    import moto

    moto.mock_ec2().start()
    return boto3.resource('ec2', region_name='eu-west-1')

@pytest.fixture()
def base_image(ec2):
    instance = ec2.create_instances(
        ImageId='ami-42424242',
        MinCount=1,
        MaxCount=1,
        InstanceType='m4.xlarge',
        EbsOptimized=True
    )[0]

    image = instance.create_image(
        Name='foo',
        Description='Foo'
    )
    return image

@pytest.fixture()
def copied_image(ec2, base_image):
    time.sleep(1) # Ensure there is a CreationDate difference
    r = runner.invoke(shipami, ['copy', base_image.id])

    image_id = r.output.strip()
    image = ec2.Image(image_id)
    return image

@pytest.fixture()
def deleted_copied_image(ec2, base_image, copied_image):
    r = runner.invoke(shipami, ['delete', '-f', base_image.id])

    return copied_image

@pytest.fixture()
def released_image(ec2, base_image):
    RELEASE = '1.0.0'
    NAME = 'foo'

    time.sleep(1) # Ensure there is a CreationDate difference
    r = runner.invoke(shipami, ['release', base_image.id, RELEASE, '--name', NAME])

    image_id = r.output.strip()
    image = ec2.Image(image_id)
    return image

class TestCli:

    def test_version(self):
        r = runner.invoke(shipami, ['--version'])

        assert r.exit_code == 0
        assert VERSION in r.output

    def test_list_unmanaged(self, base_image):
        r = runner.invoke(shipami, ['list'])

        assert r.exit_code == 0
        assert base_image.id in r.output
        assert base_image.name in r.output

    def test_list(self, base_image, released_image):
        r = runner.invoke(shipami, ['list'])

        lines = r.output.splitlines()

        assert r.exit_code == 0
        assert base_image.id in lines[2]
        assert 'origin' in lines[2]
        assert 'eu-west-1:{}'.format(released_image.id) in lines[2]

        assert released_image.id in lines[1]
        assert 'origin' not in lines[1]
        assert 'eu-west-1:{}'.format(base_image.id) in lines[1]

    def test_list_quiet(self, base_image, released_image):
        r = runner.invoke(shipami, ['list', '-q'])

        lines = r.output.splitlines()

        assert r.exit_code == 0
        assert len(lines) == 2
        assert base_image.id in lines[1]
        assert released_image.id in lines[0]

    def test_list_filter_simple(self, base_image, released_image):
        r = runner.invoke(shipami, ['list', '-f', 'release=1.0.0'])

        lines = r.output.splitlines()

        assert r.exit_code == 0
        assert len(lines) == 2
        assert ' {} '.format(released_image.id) in lines[1]

    def test_list_filter_managed(self, base_image, released_image):
        r = runner.invoke(shipami, ['list', '-f', 'managed=yes'])

        lines = r.output.splitlines()

        assert r.exit_code == 0
        assert len(lines) == 2
        assert ' {} '.format(released_image.id) in lines[1]

    def test_list_filter_not_managed(self, base_image, released_image):
        r = runner.invoke(shipami, ['list', '-f', 'managed=no'])

        lines = r.output.splitlines()

        assert r.exit_code == 0
        assert len(lines) == 2
        assert ' {} '.format(base_image.id) in lines[1]

    def test_list_filter_bad(self, base_image, released_image):
        r = runner.invoke(shipami, ['list', '-f', 'invalid'])

        assert r.exit_code == 2
        assert 'Invalid value' in r.output

    def test_list_filter_unknown(self, base_image, released_image):
        r = runner.invoke(shipami, ['list', '-f', 'invalid=42'])

        assert r.exit_code == 2
        assert 'Invalid value' in r.output

    def test_list_filter_quiet(self, base_image, released_image):
        r = runner.invoke(shipami, ['list', '-q', '-f', 'release=1.0.0'])

        lines = r.output.splitlines()

        assert r.exit_code == 0
        assert len(lines) == 1
        assert released_image.id in lines[0]

    def test_show_unmanaged(self, base_image):
        r = runner.invoke(shipami, ['show', base_image.id])

        assert r.exit_code == 0

    def test_show_managed(self, released_image):
        r = runner.invoke(shipami, ['show', released_image.id])

        assert r.exit_code == 0

    def test_show_inexistant_id(self):
        r = runner.invoke(shipami, ['show', 'ami-42424242'])

        assert r.exit_code == 1
        assert 'Error: The image id \'[ami-42424242]\' does not exist' in r.output

    def test_copy(self, ec2, base_image):
        image_number = len(ec2.meta.client.describe_images()['Images'])

        expected_tags = [
            {
                'Key': 'shipami:managed',
                'Value': 'True'
            },
            {
                'Key': 'shipami:copied_from',
                'Value': 'eu-west-1:{}'.format(base_image.id)
            }
        ]

        r = runner.invoke(shipami, ['copy', base_image.id])

        returned_image_id = r.output.strip()
        image = ec2.Image(returned_image_id)

        assert r.exit_code == 0
        assert len(ec2.meta.client.describe_images()['Images']) == image_number + 1
        assert image.name == base_image.name
        assert sorted(image.tags, key=lambda _: _['Key']) == sorted(expected_tags, key=lambda _: _['Key'])

    def test_copy_wait(self, ec2, base_image):
        image_number = len(ec2.meta.client.describe_images()['Images'])

        expected_tags = [
            {
                'Key': 'shipami:managed',
                'Value': 'True'
            },
            {
                'Key': 'shipami:copied_from',
                'Value': 'eu-west-1:{}'.format(base_image.id)
            }
        ]

        r = runner.invoke(shipami, ['copy', base_image.id, '--wait'])

        returned_image_id = r.output.strip()
        image = ec2.Image(returned_image_id)

        assert r.exit_code == 0
        assert len(ec2.meta.client.describe_images()['Images']) == image_number + 1
        assert image.name == base_image.name
        assert sorted(image.tags, key=lambda _: _['Key']) == sorted(expected_tags, key=lambda _: _['Key'])

    def test_copy_invalid_name(self, base_image):
        NAME = 'aa'

        r = runner.invoke(shipami, ['copy', base_image.id, '--name', NAME])

        assert 'Error:' in r.output

    def test_copy_cleanup_name(self, ec2, base_image):
        NAME = 'foo#bar'

        r = runner.invoke(shipami, ['copy', base_image.id, '--name', NAME])

        returned_image_id = r.output.strip()
        image = ec2.Image(returned_image_id)

        assert r.exit_code == 0
        assert image.name == 'foo-bar'

    def test_copy_inexistant_id(self, ec2, base_image):
        r = runner.invoke(shipami, ['copy', 'ami-42424242'])

        assert r.exit_code == 1
        assert 'Error: The image id \'[ami-42424242]\' does not exist' in r.output

    def test_release(self, ec2, base_image):
        RELEASE = '1.0.0'
        NAME = 'foo'
        image_number = len(ec2.meta.client.describe_images()['Images'])

        expected_tags = [
            {
                'Key': 'shipami:managed',
                'Value': 'True'
            },
            {
                'Key': 'shipami:release',
                'Value': '1.0.0'
            },
            {
                'Key': 'shipami:copied_from',
                'Value': 'eu-west-1:{}'.format(base_image.id)
            }
        ]

        r = runner.invoke(shipami, ['release', base_image.id, RELEASE, '--name', NAME])

        returned_image_id = r.output.strip()
        image = ec2.Image(returned_image_id)

        assert r.exit_code == 0
        assert len(ec2.meta.client.describe_images()['Images']) == image_number + 1
        assert image.name == '{}-{}'.format(NAME, RELEASE)
        assert sorted(image.tags, key=lambda _: _['Key']) == sorted(expected_tags, key=lambda _: _['Key'])

    def test_delete(self, ec2, copied_image):
        copied_image_id = copied_image.id
        r = runner.invoke(shipami, ['delete', copied_image_id])

        returned_image_id = r.output.strip()

        assert r.exit_code == 0
        assert len(ec2.meta.client.describe_images()['Images']) == 1
        assert returned_image_id == copied_image_id

    def test_delete_not_managed(self, ec2, base_image):
        base_image_id = base_image.id
        r = runner.invoke(shipami, ['delete', base_image_id])

        assert r.exit_code == 1
        assert len(ec2.meta.client.describe_images()['Images']) == 1
        assert 'Error: {} is either a release or not managed by shipami'.format(base_image_id) in r.output

    def test_delete_released_image(self, ec2, released_image):
        released_image_id = released_image.id
        r = runner.invoke(shipami, ['delete', released_image_id])

        assert r.exit_code == 1
        assert len(ec2.meta.client.describe_images()['Images']) == 2
        assert 'Error: {} is either a release or not managed by shipami'.format(released_image_id) in r.output

    def test_delete_force(self, ec2, base_image):
        base_image_id = base_image.id
        r = runner.invoke(shipami, ['delete', '--force', base_image_id])

        returned_image_id = r.output.strip()

        assert r.exit_code == 0
        assert len(ec2.meta.client.describe_images()['Images']) == 0
        assert returned_image_id == base_image_id

    def test_delete_force_released(self, ec2, released_image):
        released_image_id = released_image.id
        r = runner.invoke(shipami, ['delete', '--force', released_image_id])

        returned_image_id = r.output.strip()

        assert r.exit_code == 0
        assert len(ec2.meta.client.describe_images()['Images']) == 1
        assert returned_image_id == released_image_id

    def test_delete_deleted_source(self, ec2, deleted_copied_image):
        deleted_copied_image_id = deleted_copied_image.id
        r = runner.invoke(shipami, ['delete', deleted_copied_image_id])

        print(r.output)

        returned_image_id = r.output.strip()

        assert r.exit_code == 0
        assert len(ec2.meta.client.describe_images()['Images']) == 0
        assert returned_image_id == deleted_copied_image_id



from . import __version__ as VERSION

import json
import logging
import click

from tabulate import tabulate
import datetime, timeago, dateutil.parser

from shipami.core import ShipAMI

logging.basicConfig()
logger = logging.getLogger(__name__)

state_colors = {
    'available': 'green',
    'pending': 'yellow',
    'failed': 'red'
}

def validate_filter(ctx, param, filters):
    validated_filters = []
    for f in filters:
        try:
            k, v = f.split('=')
            validated_filters.append((k, v))
        except ValueError:
            raise click.BadParameter('filter must be in format "key=value"')
    return tuple(validated_filters)

@click.group()
@click.version_option(VERSION)
@click.option('--region')
@click.option('-v', '--verbose', is_flag=True, default=False)
@click.pass_context
def cli(ctx, region, verbose):
    """CLI tool to manage AWS AMI and Marketplace"""
    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    ctx.obj = ShipAMI(region)


@cli.command()
@click.option('--quiet', '-q', is_flag=True)
@click.option('filter_', '--filter', '-f', multiple=True, callback=validate_filter)
@click.option('--color/--no-color', default=True)
@click.pass_obj
def list(shipami, filter_, quiet, color):
    headers = ['NAME', 'RELEASE', 'ID', 'STATE', 'CREATED', 'MANAGED', 'COPIED FROM', 'COPIED TO']
    headers_mapping = {
        'NAME': 'Name',
        'RELEASE': 'Release',
        'ID': 'ImageId',
        'STATE': 'State',
        'CREATED': 'CreationDate',
        'MANAGED': 'Managed',
        'COPIED FROM': 'CopiedFrom',
        'COPIED TO': 'CopiedTo'
    }

    filters = ['NAME', 'RELEASE', 'ID', 'STATE', 'MANAGED']
    filters_mapping = {}
    for f in filters:
        filters_mapping[f.strip().lower()] = f

    for k, v in filter_:
        if k not in filters_mapping.keys():
            raise click.BadParameter('available filters are {}'.format(filters_mapping.keys()))

    def makefilter(k, v, attr):
        def f(_):
            x = _.get(attr)
            if x is None:
                return False
            if isinstance(x, bool):
                bool_mapping = {'yes': True, 'no': False}
                return x is bool_mapping.get(v)
            return v in x
        return f

    now = datetime.datetime.utcnow()
    try:
        images = shipami.list()
    except RuntimeError as e:
        raise click.ClickException(str(e))

    for k, v in filter_:
        attr = headers_mapping.get(filters_mapping.get(k))
        images = filter(makefilter(k, v, attr), images)

    images = sorted(images, key=lambda _: _['CreationDate'], reverse=True)

    if quiet:
        for image in images: click.echo(image.get('ImageId'))
    else:
        d = []
        for image in images:
            row = []
            for col in headers:
                value = image.get(headers_mapping.get(col))
                if col is 'STATE':
                    if color:
                        value = click.style(value, fg=state_colors.get(value))
                if col is 'CREATED':
                    value = timeago.format(dateutil.parser.parse(value, ignoretz=True), now)
                if col is 'MANAGED':
                    value = 'yes' if value else 'no'
                    if color and value is 'yes':
                        value = click.style(value, fg='white', bold=True)
                if col is 'COPIED TO':
                    if value and color:
                        value = click.style(value, fg='blue')
                if col is 'COPIED FROM':
                    if value and color:
                        value = click.style(value, fg='blue')
                    if value is None and image.get(headers_mapping.get('MANAGED')) is False:
                        value = 'origin'
                row.append(value)
            d.append(row)
        if d: print(tabulate(d, headers=headers, tablefmt='plain'))


@cli.command()
@click.argument('image-id')
@click.pass_obj
def show(shipami, image_id):
    try:
        image = shipami.show(image_id)
    except RuntimeError as e:
        raise click.ClickException(str(e))

    click.echo('id:\t{}'.format(image.get('ImageId')))
    click.echo('name:\t{}'.format(image.get('Name')))
    click.echo('state:\t{}'.format(click.style(image.get('State'), fg=state_colors.get(image.get('State')))))

    if image.get('Tags'):
        click.echo('tags:')
        for tag in sorted(image['Tags'], key=lambda _: _['Key']):
            key = tag.get('Key')
            value = tag.get('Value')
            color = 'blue' if 'shipami:' in key else 'white'
            click.echo('  {}: {}'.format(key, click.style(value, fg=color, bold=True)))

    if image.get('BlockDeviceMappings'):
        click.echo('devices mappings:')
        for block_device_mapping in image.get('BlockDeviceMappings', []):
            click.echo('  {} {}Go type:{}'.format(
                    block_device_mapping['DeviceName'],
                    block_device_mapping['Ebs']['VolumeSize'],
                    block_device_mapping['Ebs']['VolumeType']
                )
            )

    if image.get('Shares'):
        click.echo('shared with:')
        for share in image.get('Shares'):
            click.echo('  {}'.format(share['UserId']), nl=False)
            if share.get('Marketplace') is not None:
                click.echo(' (AWS MARKETPLACE)', nl=False)
                if share.get('Marketplace') is True:
                    click.secho(' OK', fg='green', nl=False)
                else:
                    click.secho(' PARTIAL', fg='yellow', nl=False)
            click.echo()


@cli.command()
@click.argument('image-id')
@click.option('--name')
@click.option('--description')
@click.option('--source-region')
@click.option('--copy-tags/--no-copy-tags', default=True)
@click.option('--copy-tags-to-snapshots/--no-copy-tags-to-snapshots', default=False)
@click.option('--copy-permissions/--no-copy-permissions', default=False)
@click.option('--wait/--no-wait', default=False)
@click.pass_obj
def copy(shipami, **kwargs):
    try:
        image_id = shipami.copy(kwargs.pop('image_id'), **kwargs)
    except RuntimeError as e:
        raise click.ClickException(str(e))

    click.echo(image_id)


@cli.command()
@click.argument('image-id')
@click.argument('release')
@click.option('--name')
@click.option('--description')
@click.option('--source-region')
@click.option('--copy-tags/--no-copy-tags', default=True)
@click.option('--copy-tags-to-snapshots/--no-copy-tags-to-snapshots', default=False)
@click.option('--copy-permissions/--no-copy-permissions', default=False)
@click.option('--wait/--no-wait', default=False)
@click.pass_obj
def release(shipami, **kwargs):
    try:
        image_id = shipami.release(kwargs.pop('image_id'), kwargs.pop('release'), **kwargs)
    except RuntimeError as e:
        raise click.ClickException(str(e))

    click.echo(image_id)


@cli.command()
@click.argument('image-id')
@click.option('--account-id')
@click.option('--remove', is_flag=True, default=False)
@click.pass_obj
def share(shipami, **kwargs):
    try:
        shipami.share(kwargs.pop('image_id'), **kwargs)
    except RuntimeError as e:
        raise click.ClickException(str(e))


@cli.command()
@click.argument('image-id', nargs=-1)
@click.option('--force', '-f', is_flag=True, default=False)
@click.pass_obj
def delete(shipami, image_id, force):
    try:
        deleted = shipami.delete(image_id, force)
    except RuntimeError as e:
        raise click.ClickException(str(e))

    for d in deleted:
        click.echo(d)



import json
import logging
import boto3
import botocore
import time

import botocore.vendored.requests.packages.urllib3 as urllib3
urllib3.disable_warnings(urllib3.exceptions.SecurityWarning)

logging.basicConfig()
logger = logging.getLogger('shipami.cli')


class ShipAMI(object):

    MARKETPLACE_REGION = 'us-east-1'
    MARKETPLACE_ACCOUNT_ID = '679593333241'

    def __init__(self, region=None):
        self._region = region or boto3.session.Session().region_name
        self._sessions = {}

    def __get_session(self, region=None):
        region = region or self._region
        session = self._sessions.get(region)
        if not session:
            self._sessions[region] = boto3.session.Session(region_name=region)
            session = self._sessions[region]
        return session

    def validate_ami_name(self, name, clean=False):
        allowed = ['(', ')', '[', ']', ' ', '.', '/', '-', '\'', '@', '_']

        if len(name) < 3 or len(name) > 128:
            raise RuntimeError('AMI Name must be 3-128 long (got: "{}")'.format(name))

        if clean:
            return ''.join(map(lambda _: _ if _.isalnum() or _ in allowed else '-', name))
        return name

    def list(self):
        result = []
        copied_keys = ['ImageId', 'Name', 'State', 'CreationDate']
        ec2 = self.__get_session().client('ec2')

        try:
            r = ec2.describe_images(
                Owners=[
                    'self'
                ]
            )
        except botocore.exceptions.ClientError as e:
            message = e.response['Error']['Message']
            logger.error(message)
            raise RuntimeError(message)

        images = r.get('Images', [])
        for image in images:
            i = {}
            for key in copied_keys:
                i[key] = image.get(key)

            i['Managed'] = self.__is_managed(image)
            i['Release'] = self.__get_tag(image, 'shipami:release')
            i['CopiedFrom'] = self.__get_tag(image, 'shipami:copied_from')
            i['CopiedTo'] = self.__get_tag(image, 'shipami:copied_to')

            result.append(i)

        return result

    def show(self, image_id):
        ec2 = self.__get_session().client('ec2')
        image = self.__get_session().resource('ec2').Image(image_id)

        try:
            r = ec2.describe_images(
                Owners=[
                    'self'
                ],
                ImageIds=[
                    image.id
                ]
            )
        except botocore.exceptions.ClientError as e:
            message = e.response['Error']['Message']
            logger.error(message)
            raise RuntimeError(message)

        try:
            result_image = r.get('Images', [])[0]
        except IndexError:
            message = 'Something went wrong'
            logger.error(message)
            raise RuntimeError(message)

        result_image['Shares'] = self.__get_image_permissions(image)
        for share in result_image['Shares']:
            if share.get('UserId') == self.MARKETPLACE_ACCOUNT_ID:
                share['Marketplace'] = self.__is_ami_shared(image)

        return result_image

    def copy(self, image_id, **kwargs):
        src_image = self.__get_session(kwargs.pop('source_region')).resource('ec2').Image(image_id)
        dst_image = self.__copy_image(src_image, **kwargs)
        return dst_image.id

    def release(self, image_id, release, **kwargs):
        image = self.__get_session().resource('ec2').Image(self.copy(image_id, name_suffix=release, **kwargs))
        self.__set_tag(image, 'shipami:release', release)
        return image.id

    def share(self, image_id, account_id=None, remove=False):
        image = self.__get_session().resource('ec2').Image(image_id)
        account_id = account_id or self.MARKETPLACE_ACCOUNT_ID
        operation = 'add' if not remove else 'remove'
        operation_log = 'adding' if not remove else 'removing'

        logger.debug('{} permissions for {} on image {}'.format(operation_log, account_id, image.id))
        self.__wait_for_image(image)
        self.__share_modify_attribute(image, 'launchPermission', operation, account_id)
        for snapshot in self.__get_image_snapshots(image):
            logger.debug('{} permissions for {} on snapshot {}'.format(operation_log, account_id, snapshot.id))
            self.__wait_for_snapshot(snapshot)
            self.__share_modify_attribute(snapshot, 'createVolumePermission', operation, account_id)

    def delete(self, image_ids, force=False):
        ec2 = self.__get_session().resource('ec2')
        deleted = []

        for image_id in image_ids:
            image = ec2.Image(image_id)

            managed = self.__is_managed(image)
            release = self.__is_release(image)

            if (not managed or release) and (not force):
                message = '{} is either a release or not managed by shipami, you must use -f to delete this image'.format(image.id)
                raise RuntimeError(message)

            copied_from = None
            if managed:
                copied_from = self.__get_tag(image, 'shipami:copied_from')
                if copied_from:
                    from_image = self.__get_copied_from_image(copied_from)
                    remove_copied_to = self.__generate_copy_tag(image)

            try:
                snapshots = self.__get_image_snapshots(image)
                logger.debug('deregistering {}'.format(image.id))
                image.deregister()
                for snapshot in snapshots:
                    logger.debug('deleting {}'.format(snapshot.id))
                    snapshot.delete()
            except botocore.exceptions.ClientError as e:
                message = e.response['Error']['Message']
                logger.error(message)
                raise RuntimeError(message)

            if managed and copied_from:
                self.__remove_copied_to(from_image, remove_copied_to)

            deleted.append(image_id)
        return deleted

    def __copy_image(self, src_image, name=None, name_suffix=None, description=None, copy_tags=True, copy_tags_to_snapshots=False, copy_permissions=False, wait=False):
        try:
            name = name or src_image.name
            if name_suffix:
                name = '-'.join([name, name_suffix])
            name = self.validate_ami_name(name, clean=True)
            description = description or src_image.description
            src_region = src_image.meta.client.meta.region_name
        except botocore.exceptions.ClientError as e:
            message = e.response['Error']['Message']
            logger.error(message)
            raise RuntimeError(message)

        try:
            logger.debug('copying image {} from {} to {}'.format(src_image.id, src_region, self._region))
            r = self.__get_session().client('ec2').copy_image(
                SourceRegion=src_region,
                SourceImageId=src_image.id,
                Name=name,
                Description=description
            )
        except botocore.exceptions.ClientError as e:
            message = e.response['Error']['Message']
            logger.error(message)
            raise RuntimeError(message)

        dst_image = self.__get_session().resource('ec2').Image(r['ImageId'])
        self.__append_tag(src_image, 'shipami:copied_to', '{}:{}'.format(self._region, dst_image.id))

        if copy_tags:
            self.__copy_tags(src_image, dst_image, copy_tags_to_snapshots)
            # removes irrelevant 'copied_to' tag
            # TODO: Find a way to filter tags properly when copying
            self.__delete_tag(dst_image, 'shipami:copied_to')

        self.__set_managed(dst_image)
        self.__set_tag(dst_image, 'shipami:copied_from', '{}:{}'.format(src_region, src_image.id))

        if copy_permissions:
            try:
                self.__wait_for_image(dst_image)
                for permission in self.__get_image_permissions(src_image):
                    account_id = permission.get('UserId')
                    logger.debug('adding launchPermission permission for {} on image {}'.format(account_id, dst_image.id))
                    self.__share_modify_attribute(dst_image, 'launchPermission', 'add', account_id)

                src_block_devices = self.__get_image_block_devices(src_image)
                for dst_block_device in self.__get_image_block_devices(dst_image):
                    dst_snapshot = dst_block_device.get('Snapshot')
                    self.__wait_for_snapshot(dst_snapshot)
                    for src_block_device in src_block_devices:
                        if src_block_device.get('DeviceName') == dst_block_device.get('DeviceName'):
                            src_snapshot = src_block_device.get('Snapshot')
                            logger.debug('found matching DeviceName for {} and {}'.format(src_snapshot.id, dst_snapshot.id))
                            for permission in self.__get_snapshot_permissions(src_snapshot):
                                account_id = permission.get('UserId')
                                if account_id == 'aws-marketplace':
                                    account_id = self.MARKETPLACE_ACCOUNT_ID
                                logger.debug('adding createVolumePermission permission for {} on snapshot {}'.format(account_id, dst_snapshot.id))
                                self.__share_modify_attribute(dst_snapshot, 'createVolumePermission', 'add', account_id)
            except botocore.exceptions.ClientError as e:
                message = e.response['Error']['Message']
                logger.error(message)
                raise RuntimeError(message)

        if wait and not copy_permissions:
            self.__wait_for_image(dst_image)

        return dst_image

    def __copy_tags(self, src_image, dst_image, copy_to_snapshots=False):
        logger.debug('copying tags from image {} to image {}'.format(src_image.id, dst_image.id))

        dst_image.create_tags(Tags=src_image.tags)
        if copy_to_snapshots:
            for snapshot in self.__get_image_snapshots(dst_image):
                logger.debug('copying tags to snapshot {}'.format(snapshot.id))
                snapshot.create_tags(Tags=src_image.tags)

    def __share_modify_attribute(self, obj, attribute, operation, account_id):
        try:
            obj.modify_attribute(
                Attribute=attribute,
                OperationType=operation,
                UserIds=[
                    account_id
                ]
            )
        except botocore.exceptions.ClientError as e:
            message = e.response['Error']['Message']
            logger.error(message)
            raise RuntimeError(message)

    def __get_copied_from_image(self, copied_from):
        try:
            region, image_id = copied_from.split(':')
        except ValueError as e:
            logger.error(e)
            raise RuntimeError(e)
        image = self.__get_session(region).resource('ec2').Image(image_id)
        return image

    def __generate_copy_tag(self, image):
        return '{}:{}'.format(self.__get_image_region(image), image.id)

    def __remove_copied_to(self, image, to_remove):
        try:
            copied_to = self.__get_tag(image, 'shipami:copied_to')

            logger.debug('removing "{}" from {} shipami:copied_to tag'.format(to_remove, image.id))
            logger.debug('shipami:copied_to: {}'.format(copied_to))

            if copied_to:
                copied_to = copied_to.split(',')
                copied_to = filter(lambda _: _ != to_remove, copied_to)
                copied_to = ','.join(copied_to)

                if copied_to:
                    logger.debug('set shipami:copied_to: {}'.format(copied_to))
                    self.__set_tag(image, 'shipami:copied_to', copied_to)
                else:
                    logger.debug('removed shipami:copied_to')
                    self.__delete_tag(image, 'shipami:copied_to')
        except RuntimeError as e:
            logger.debug(str(e))

    def __get_image_region(self, image):
        return image.meta.client.meta.region_name

    def __get_image_snapshots(self, image):
        region_name = self.__get_image_region(image)
        ec2 = self.__get_session(region_name).resource('ec2')
        snapshots = []

        # We must wait for the image to be avaiale in order to get the SnapshotIds
        self.__wait_for_image(image)
        for block_device_mapping in image.block_device_mappings:
            if block_device_mapping.get('Ebs'):
                snapshots.append(ec2.Snapshot(block_device_mapping['Ebs']['SnapshotId']))
        return snapshots

    def __get_image_block_devices(self, image):
        region_name = self.__get_image_region(image)
        ec2 = self.__get_session(region_name).resource('ec2')
        block_devices = []

        # We must wait for the image to be avaiale in order to get the SnapshotIds
        self.__wait_for_image(image)
        for block_device_mapping in image.block_device_mappings:
            if block_device_mapping.get('Ebs'):
                block_devices.append({'DeviceName': block_device_mapping.get('DeviceName'), 'Snapshot': ec2.Snapshot(block_device_mapping['Ebs']['SnapshotId'])})
        return block_devices

    def __set_managed(self, image):
        self.__set_tag(image, 'shipami:managed', 'True')

    def __append_tag(self, obj, key, value):
        p_value = self.__get_tag(obj, key)
        if p_value:
            value = ','.join([p_value, value])

        self.__set_tag(obj, key, value)

    def __get_tag(self, obj, key):
        try:
            tags = obj.tags or []
        except AttributeError:
            try:
                tags = obj.get('Tags', [])
            except AttributeError:
                return None
        except botocore.exceptions.ClientError as e:
            message = e.response['Error']['Message']
            raise RuntimeError(message)

        for tag in tags:
            if tag.get('Key') == key:
                return tag.get('Value')
        return None

    def __set_tag(self, obj, key, value):
        try:
            obj.create_tags(
                Tags=[
                    {
                        'Key': key,
                        'Value': value
                    }
                ]
            )
        except botocore.exceptions.ClientError as e:
            message = e.response['Error']['Message']
            logger.error(message)
            raise RuntimeError(message)

    def __delete_tag(self, obj, key):
        try:
            obj.meta.client.delete_tags(
                Resources=[
                    obj.id
                ],
                Tags=[
                    {
                        'Key': key
                    }
                ]
            )
        except botocore.exceptions.ClientError as e:
            message = e.response['Error']['Message']
            logger.error(message)
            raise RuntimeError(message)

    def __is_managed(self, image):
        if self.__get_tag(image, 'shipami:managed') == 'True':
            return True
        return False

    def __is_release(self, image):
        return True if self.__get_tag(image, 'shipami:release') else False

    def __get_image_permissions(self, image):
        try:
            r = image.describe_attribute(
                Attribute='launchPermission'
            )
        except botocore.exceptions.ClientError as e:
            message = e.response['Error']['Message']
            logger.error(message)
            raise RuntimeError(message)

        return r.get('LaunchPermissions', [])

    def __get_snapshot_permissions(self, snapshot):
        try:
            r = snapshot.describe_attribute(
                Attribute='createVolumePermission'
            )
        except botocore.exceptions.ClientError as e:
            message = e.response['Error']['Message']
            logger.error(message)
            raise RuntimeError(message)

        return r.get('CreateVolumePermissions', [])

    def __is_ami_shared(self, image, account_id=None):
        account_id = account_id or self.MARKETPLACE_ACCOUNT_ID

        for snapshot in self.__get_image_snapshots(image):
            if not self.__is_snapshot_shared(snapshot, account_id):
                return False
        return self.__is_image_shared(image, account_id)

    def __is_image_shared(self, image, account_id=None):
        account_id = account_id or self.MARKETPLACE_ACCOUNT_ID

        for permission in self.__get_image_permissions(image):
            if (permission.get('UserId') == account_id) or (account_id == self.MARKETPLACE_ACCOUNT_ID and permission.get('UserId') == 'aws-marketplace'):
                return True

    def __is_snapshot_shared(self, snapshot, account_id=None):
        account_id = account_id or self.MARKETPLACE_ACCOUNT_ID

        for permission in self.__get_snapshot_permissions(snapshot):
            if (permission.get('UserId') == account_id) or (account_id == self.MARKETPLACE_ACCOUNT_ID and permission.get('UserId') == 'aws-marketplace'):
                return True
        return False

    def __wait_for_image(self, image, state='available'):
        logger.debug('waiting for image {} to be {}'.format(image.id, state))
        image.wait_until_exists(
            Filters=[
                {
                    'Name': 'state',
                    'Values': [
                        state
                    ]
                }
            ]
        )
        image.reload()

    def __wait_for_snapshot(self, snapshot):
        logger.debug('waiting for snapshot {} to be ready'.format(snapshot.id))
        snapshot.wait_until_completed(
            Filters=[
                {
                    'Name': 'status',
                    'Values': [
                        'completed',
                        'error'
                    ]
                }
            ]
        )
        snapshot.reload()

    def __wait_for_block_devices(self, image):
        logger.debug('waiting for block devices')
        while not image.block_device_mappings:
            image.reload()
            time.sleep(1)



from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

from shipami import __version__

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='shipami',

    version=__version__,

    description='CLI tool to manage AWS AMI and Marketplace',
    long_description=long_description,

    url='http://github.com/wnkz/shipami',
    download_url='https://github.com/wnkz/shipami/archive/{}.tar.gz'.format(__version__),

    author='wnkz',
    author_email='wnkz@users.noreply.github.com',

    license='MIT',

    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',

        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    keywords='aws ec2 ami marketplace',

    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),

    install_requires=[
        'click==6.7',
        'botocore>=1.5.0,<1.6.0',
        'boto3>=1.4.4',
        'tabulate>=0.7.7',
        'timeago>=1.0.7'
    ],

    entry_points={
        'console_scripts': [
            'shipami=shipami.cli:cli',
        ],
    },

    setup_requires=[
        'pytest-runner>=2,<3dev'
    ],

    tests_require=[
        'pytest>=3.0.6',
        'moto>=0.4.31'
    ],

    zip_safe=False
)
